#define GRIDDIM 1
#define WORLDDIM 2

#define ALBERTAGRID // the mesh file must be oriented consistently since it does not perform any check
#define POLORDER 1

#define USE_ISTL 1
#define USE_MCF 1

#include "config.h"
#include <dune/common/parallel/mpicollectivecommunication.hh>
#include <dune/common/parallel/mpihelper.hh>
#include <dune/grid/io/file/gmshreader.hh>
#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/io/parameter.hh>

#include <dune/fem/gridpart/leafgridpart.hh>
#include <dune/common/timer.hh>
#include <dune/fem/space/common/functionspace.hh>
#include <dune/fem/space/lagrange.hh>

#if USE_ISTL
#include <dune/fem/function/blockvectorfunction.hh>
#else
#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/operator/linear/spoperator.hh>
#endif

#include <iostream>
#include "surfacelaplace.hh"
#include "productnormalNEW.hh"

int main(int argc, char** argv)
{
  try{
	Dune::Fem::MPIManager::initialize(argc,argv);
	Dune::Fem::Parameter::append(argc,argv);
	Dune::Fem::Parameter::append((argc<2)?("/home/ma/a/as7211/dune-2.3.1/dune-schur/src/parameter"):(argv[1]));
	Dune::CollectiveCommunication<MPI_Comm> collCom(MPI_COMM_WORLD);

	// get size and rank
	//int nProcs(collCom.size());
	int rank(collCom.rank());

	typedef Dune::GridSelector::GridType HostGridType;
        typedef Dune::Fem::LeafGridPart<HostGridType> HostGridPartType;
	const std::string FileName(Dune::Fem::Parameter::getValue<std::string>("FileName","mesh.msh"));
	typedef Dune::GridFactory<HostGridType> GridFactory;
	GridFactory gridFactory;
	Dune::GmshReader<HostGridType>::read(gridFactory,FileName);
	Dune::shared_ptr<HostGridType> gridPtr(gridFactory.createGrid());
	HostGridType& grid(*gridPtr);
    	grid.loadBalance();

	// spaces
	typedef Dune::Fem::FunctionSpace<double,double,HostGridType::dimensionworld,1> ScalarFunctionSpaceType;
	typedef Dune::Fem::FunctionSpace<double,double,HostGridType::dimensionworld,HostGridType::dimensionworld> VectorFunctionSpaceType;
	typedef Dune::Fem::LagrangeDiscreteFunctionSpace<ScalarFunctionSpaceType,HostGridPartType,1> DiscreteScalarFunctionSpaceType;
	typedef Dune::Fem::LagrangeDiscreteFunctionSpace<VectorFunctionSpaceType,HostGridPartType,1> DiscreteVectorFunctionSpaceType;

	#if USE_ISTL
	typedef Dune::Fem::ISTLBlockVectorDiscreteFunction<DiscreteScalarFunctionSpaceType> ScalarDiscreteFunctionType;
	typedef Dune::Fem::ISTLBlockVectorDiscreteFunction<DiscreteVectorFunctionSpaceType> VectorDiscreteFunctionType;
	typedef Dune::Fem::ISTLLinearOperator<ScalarDiscreteFunctionType,ScalarDiscreteFunctionType> ScalarLinearOperatorImp;
        typedef Dune::Fem::ISTLLinearOperator<VectorDiscreteFunctionType,VectorDiscreteFunctionType> VectorLinearOperatorImp;
        typedef Dune::Fem::ISTLLinearOperator<ScalarDiscreteFunctionType,VectorDiscreteFunctionType> ScalarVectorLinearOperatorImp;
        typedef Dune::Fem::ISTLLinearOperator<VectorDiscreteFunctionType,ScalarDiscreteFunctionType> VectorScalarLinearOperatorImp;
	#else
        typedef Dune::Fem::AdaptiveDiscreteFunction<DiscreteScalarFunctionSpaceType> ScalarDiscreteFunctionType;
        typedef Dune::Fem::AdaptiveDiscreteFunction<DiscreteVectorFunctionSpaceType> VectorDiscreteFunctionType;
        typedef Dune::Fem::SparseRowLinearOperator<ScalarDiscreteFunctionType,ScalarDiscreteFunctionType> ScalarLinearOperatorImp;
        typedef Dune::Fem::SparseRowLinearOperator<VectorDiscreteFunctionType,VectorDiscreteFunctionType> VectorLinearOperatorImp;
        typedef Dune::Fem::SparseRowLinearOperator<ScalarDiscreteFunctionType,VectorDiscreteFunctionType> ScalarVectorLinearOperatorImp;
        typedef Dune::Fem::SparseRowLinearOperator<VectorDiscreteFunctionType,ScalarDiscreteFunctionType> VectorScalarLinearOperatorImp;
	#endif

	// operators

	// typedef Dune::SurfaceLaplaceFEOp<ScalarDiscreteFunctionType,ScalarLinearOperatorImp> ScalarLaplaceOperator;
        // typedef Dune::SurfaceLaplaceFEOp<VectorDiscreteFunctionType,VectorLinearOperatorImp> VectorLaplaceOperator;
	typedef Dune::ProductNormalFEOp<ScalarDiscreteFunctionType,VectorDiscreteFunctionType,ScalarVectorLinearOperatorImp> CouplingScalarVectorOperator;
        typedef Dune::ProductNormalFEOp<VectorDiscreteFunctionType,ScalarDiscreteFunctionType,VectorScalarLinearOperatorImp> CouplingVectorScalarOperator;

	// leaf grid, spaces and f.e. functions
	HostGridPartType gridPart(grid);
	DiscreteScalarFunctionSpaceType scalarSpace(gridPart);
        DiscreteVectorFunctionSpaceType vectorSpace(gridPart);
	VectorDiscreteFunctionType rhsPosition("rhsPosition",vectorSpace);
        ScalarDiscreteFunctionType rhsCurvature("rhsCurvature",scalarSpace);
        VectorDiscreteFunctionType solutionPosition("solutionPosition",vectorSpace);
        ScalarDiscreteFunctionType solutionCurvature("solutionCurvature",scalarSpace);
	rhsPosition.clear(); rhsCurvature.clear(); solutionPosition.clear(); solutionCurvature.clear();

	// assemble the four operators
	// ScalarLaplaceOperator scalarOp (scalarSpace);
	// scalarOp.assemble();
        // VectorLaplaceOperator vectorOp (vectorSpace);
        // vectorOp.assemble();
	CouplingScalarVectorOperator scalarvectorOp(scalarSpace,vectorSpace);
	
	return 0;
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  }
}
