#ifndef DUNE_PRODUCTNORMALNEW_HH
#define DUNE_PRODUCTNORMALNEW_HH

#include <dune/common/timer.hh>
#include <dune/fem/quadrature/cachingquadrature.hh>
#include <dune/fem/operator/common/operator.hh>
#include <dune/fem/operator/common/stencil.hh>

namespace Dune
{
	template< class DiscreteDomainFunction, class DiscreteRangeFunction, class LinearOperatorImp >
	class ProductNormalFEOp : public Fem::Operator <DiscreteDomainFunction, DiscreteRangeFunction> 
	{

	public:
		typedef DiscreteDomainFunction DiscreteDomainFunctionType;
		typedef DiscreteRangeFunction DiscreteRangeFunctionType;
		typedef LinearOperatorImp LinearOperatorType;	
		typedef ProductNormalFEOp<DiscreteDomainFunctionType,DiscreteRangeFunctionType,LinearOperatorType> ThisType;
		
		typedef typename DiscreteDomainFunctionType :: DiscreteFunctionSpaceType DiscreteDomainFunctionSpaceType;
		typedef typename DiscreteRangeFunctionType :: DiscreteFunctionSpaceType DiscreteRangeFunctionSpaceType;
		
		typedef typename DiscreteDomainFunctionSpaceType :: IteratorType IteratorType;
		typedef typename DiscreteDomainFunctionSpaceType :: RangeFieldType RangeFieldType;
		
		typedef typename DiscreteDomainFunctionSpaceType :: GridPartType GridPartType;
		typedef typename DiscreteDomainFunctionSpaceType :: GridType GridType;
		typedef typename GridType :: template Codim < 0 > :: Geometry :: GlobalCoordinate CoordinateType;
		typedef typename GridPartType::template Codim< 0 >::IteratorType GridIteratorType;
		typedef typename IteratorType :: Entity EntityType;
		typedef typename EntityType :: Geometry GeometryType;
		typedef typename GridType::ctype ctype;
		
	protected:

		typedef typename DiscreteDomainFunctionSpaceType :: BasisFunctionSetType DomainBaseFunctionSetType;
		typedef typename DiscreteRangeFunctionSpaceType :: BasisFunctionSetType RangeBaseFunctionSetType;

		typedef typename DiscreteDomainFunctionSpaceType :: RangeType DomainType;
		typedef typename DiscreteRangeFunctionSpaceType :: RangeType RangeType;
		
		enum{dimRange = GridType::dimension};
		
	public:

		// type of quadrature to be used
		typedef Fem::CachingQuadrature < GridPartType, 0 > QuadratureType;
		
		// get important types from the MatrixObject 
		typedef typename LinearOperatorType :: LocalMatrixType LocalMatrixType;
		typedef typename LinearOperatorType :: MatrixType MatrixType;
		
		// constructor 
		explicit ProductNormalFEOp ( const DiscreteDomainFunctionSpaceType & domainSpace, const DiscreteRangeFunctionSpaceType & rangeSpace)
		: domainspace_( domainSpace ),rangespace_( rangeSpace ),oper_( "productnormaloper", domainspace_,rangespace_ )
		{}
        
		ProductNormalFEOp ( const ThisType & ) = delete;
		
		// apply the operator
		virtual inline void operator() ( const DiscreteDomainFunctionType & u, DiscreteRangeFunctionType & w ) const 
		{ 
			oper_.apply( u, w ); 
		}                                                               
		
		// print the system matrix into a stream
		void print ( std::ostream & out = std::cout ) const 
		{ 
			oper_.matrix().print( out ); 
		}
		
		inline LinearOperatorType & getOperator () const
		{ 
			return oper_; 
		}

		inline const DiscreteDomainFunctionSpaceType& domainSpace() const {return domainspace_;}
		inline const DiscreteRangeFunctionSpaceType& rangeSpace() const {return rangespace_;}
   		
		// perform a grid walkthrough and assemble the global matrix
		void assemble ()  const
		{
      			// reserve memory for matrix
      			typedef Fem::DiagonalAndNeighborStencil<DiscreteDomainFunctionSpaceType,DiscreteRangeFunctionSpaceType> StencilType;
			StencilType stencil(domainspace_,rangespace_);
      			oper_.reserve(stencil);
			Timer timer;
			oper_.clear();
			
			// apply local matrix assembler on each element
			const IteratorType itEnd(domainspace_.end());
			for(IteratorType it = domainspace_.begin(); it != itEnd; ++it)
				assembleLocal( *it , get_unit_normal ( * it )  );
			
			// get elapsed time 
			const double assemblyTime = timer.elapsed();
			std :: cout << "Time to assemble our rectangular matrix: " << assemblyTime << " seconds." << std :: endl;
		}
		
	protected:

               template<typename R,int dim>
                struct Selector{};

                template<typename R>
                struct Selector<R,1>
                {
                        template<typename P,typename C,typename N>
                        R operator()(const R& weight,const P& range,const C& domain,const N& normal,const size_t& i, const size_t& j)
                        {
                                return weight*domain[j]*(range[i]*normal);
                        }
                };

                template<typename R>
                struct Selector<R,WORLDDIM>
                {
                        template<typename P,typename C,typename N>
                        R operator()(const R& weight,const P& range,const C& domain,const N& normal,const size_t& i, const size_t& j)
                        {
                                return weight*range[i]*(domain[j]*normal);
                        }
                };

		// assemble local matrix for given entity
		template < class EntityType >
		void assembleLocal( const EntityType & entity , const CoordinateType & normal ) const 
		{
			// assert that matrix is not build on ghost elements 
			assert( entity.partitionType() != GhostEntity );
			
			// cache geometry of entity 
			const GeometryType & geometry = entity.geometry();
			
			// get local matrix from matrix object
			LocalMatrixType localMatrix = oper_.localMatrix ( entity, entity );
			
			// get base function set
			const DomainBaseFunctionSetType & domainBaseSet = localMatrix.domainBasisFunctionSet();
			const RangeBaseFunctionSetType & rangeBaseSet = localMatrix.rangeBasisFunctionSet();
			
			// create quadrature of appropriate order
			const typename DiscreteDomainFunctionSpaceType::LagrangePointSetType& pointSet(domainspace_.lagrangePointSet(entity));
			
			// loop over all quadrature points
			const size_t numQuadraturePoints = pointSet.nop();

			const size_t domainBlockSize(domainspace_.localBlockSize);
                        const size_t rangeBlockSize(rangespace_.localBlockSize);
			
			for( size_t pt = 0; pt !=numQuadraturePoints; ++pt)
			{
				// get local coordinate of quadrature point 
				const typename DiscreteDomainFunctionSpaceType::LagrangePointSetType:: CoordinateType& x(pointSet.point(pt));
				std::vector<DomainType> domainValues(domainspace_.blockMapper().maxNumDofs()*domainBlockSize);
                		std::vector<RangeType> rangeValues(rangespace_.blockMapper().maxNumDofs()*rangeBlockSize);
				domainBaseSet.evaluateAll(x,domainValues);
				rangeBaseSet.evaluateAll(x,rangeValues);
		
				RangeFieldType weight = geometry.integrationElement( x ) / numQuadraturePoints ;
				
				const size_t rows    = localMatrix.rows();
                        	const size_t columns = localMatrix.columns();

                        	for( size_t i = 0; i < rows; i++ ) // # of rows = # DoFs for range
                        	{
                                	for ( size_t j = 0; j < columns; j++ ) // # of columns = # DoFs for domain
                                	{
                                        	RangeFieldType value = 0.;
                                        	constexpr int dim(DomainType::dimension);
                                        	Selector<RangeFieldType,dim> selector;
                                        	value=selector(weight,rangeValues,domainValues,normal,i,j);
                                        	localMatrix.add( i, j, value );
                                	}	
                        	}

			}
		}

		
	template < class EntityType >
	const CoordinateType get_unit_normal ( const EntityType & entity ) const
		{			
			// cache geometry of entity 
			const GeometryType & geometry = entity.geometry();
			CoordinateType normal;	
			#if WORLDDIM == 3
				CoordinateType vertex0 = geometry.corner(0);
				CoordinateType factor1 = geometry.corner(1) - vertex0;
				CoordinateType factor2 = geometry.corner(2) - vertex0;
			
				normal[0] = factor1[1] * factor2[2] - factor1[2] * factor2[1];
				normal[1] = factor1[2] * factor2[0] - factor1[0] * factor2[2];
				normal[2] = factor1[0] * factor2[1] - factor1[1] * factor2[0];
				normal /= ( normal.two_norm() );
			#else 
					CoordinateType diff = geometry.corner(1) - geometry.corner(0);
					normal[0] = (-1) * diff[1];
					normal[1] = diff[0];
					normal /= ( normal.two_norm() );
			#endif

			return normal;
		}

	protected:
		
		const DiscreteDomainFunctionSpaceType & domainspace_;
		const DiscreteRangeFunctionSpaceType & rangespace_;
		mutable LinearOperatorType oper_;
	};                                                                  
	
}

#endif
